"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var express = require("express");
var nunjucks = require("nunjucks");
var fs = require("fs");
require("reflect-metadata");
var typeorm_1 = require("typeorm");
var session = require("express-session");
var bodyParser = require("body-parser");
var multer = require("multer");
var cors = require("cors");
var apollo_server_express_1 = require("apollo-server-express");
var expressGenerator = require("./express-generator");
var graphqlGenerator = require("./graphql/generator");
var setup_1 = require("./entities/setup");
var jsonwebtoken_1 = require("jsonwebtoken");
var translations = {};
var routes = {};
Object.defineProperty(Array.prototype, "serialize", {
    enumerable: false,
    value: function () {
        var r = [];
        for (var _i = 0, _a = this; _i < _a.length; _i++) {
            var el = _a[_i];
            if (el.serialize) {
                r.push(el.serialize());
            }
            else {
                r.push(el);
            }
        }
        return r;
    }
});
Object.defineProperty(Array.prototype, "addHiddenField", {
    enumerable: false,
    value: function (field) {
        for (var _i = 0, _a = this; _i < _a.length; _i++) {
            var el = _a[_i];
            if (el.addHiddenField) {
                el.addHiddenField(field);
            }
        }
    }
});
Object.defineProperty(Array.prototype, "removeHiddenField", {
    enumerable: false,
    value: function (field) {
        for (var _i = 0, _a = this; _i < _a.length; _i++) {
            var el = _a[_i];
            if (el.removeHiddenField) {
                el.removeHiddenField(field);
            }
        }
    }
});
/**
 * Utility function to check if we are in the production environment.
 * @return true if the NODE_ENV is set to "production", false otherwise
 */
function isProduction() {
    return process.env.NODE_ENV === "production";
}
exports.isProduction = isProduction;
/**
 * Implementation of the tr filer for the nunjucks engine.
 * It trying to understand the current language from the request. The fallback
 * uses the defaultLanguage set on the app.
 */
function translate(str) {
    try {
        var lang = this.getVariables()["lang"];
        if (!lang) {
            var req = this.ctx.req;
            lang = req.acceptsLanguages()[0];
            if (lang === "*") {
                lang = null;
            }
        }
        if (!lang) {
            var app = this.ctx.req.app.get("app");
            lang = app.config.defaultLanguage;
        }
        return performTranslation(str, translations[lang]);
    }
    catch (e) {
        console.log(e);
        console.log(this);
    }
    return str;
}
function performTranslation(str, translations) {
    var translation = translations[str];
    if (translation) {
        return translation;
    }
    var start = str.indexOf("{{");
    var end = str.indexOf("}}");
    if (start != -1 && end != -1) {
        var key = str.substring(start + 2, end);
        translation = translations[key.trim()];
        return str.replace("{{" + key + "}}", translation);
    }
    return str;
}
/**
 * Apply parameters to an URL. If a parameter is not found as path parameter,
 * it is added as query parameter.
 * @param url the url to compile
 * @param parameters a plain object containing the parameters
 * @return the compiled url
 */
function applyParameterstoUrl(url, parameters) {
    if (!parameters) {
        return url;
    }
    if (url.indexOf("?") == -1) {
        url += "?";
    }
    else {
        url += "&";
    }
    for (var key in parameters) {
        if (url.indexOf(":" + key) == -1) {
            url += key + "=" + parameters[key] + "&";
        }
        else {
            url = url.replace(":" + key, parameters[key]);
        }
    }
    if (url.endsWith("?") || url.endsWith("&")) {
        url = url.substring(0, url.length - 1);
    }
    return url;
}
/**
 * Transform the route name to a URL and compile it with the given parameters.
 * @param name the route name (or, eventually, the path)
 * @param parameters a plain object containing the parameters
 * @return the compiled url
 */
function route(name, parameters) {
    var url = name;
    if (routes[name]) {
        url = routes[name];
    }
    return applyParameterstoUrl(url, parameters);
}
/**
 * Implementation of the old filter function. This function returns the previous
 * value of the input form. Fallback to the defaultValue.
 * @param name the name used in the form
 * @param defaultValue a fallback value
 * @return the previous value or defaultValue
 */
function old(name, defaultValue) {
    var req = this.ctx.req;
    if (req.body && req.body[name]) {
        return req.body[name];
    }
    if (req.query[name]) {
        return req.query[name];
    }
    return defaultValue;
}
/**
 * Implementation of the format filter function to format a float number.
 * By default, the number is formatted with 2 decimal numbers.
 * @param val the number to format
 * @param decimal the number of decimal number
 * @return the formatted number as a string
 */
function format(val, decimal) {
    if (decimal === void 0) { decimal = 2; }
    return Number(val).toFixed(decimal);
}
/**
 * Implementation of the resolvePath global function. Using this function, it is
 * possible to refer to any views with a virtual folder containing all the available
 * views.
 * @param path the virtual absolute path of the view
 * @return the absolute path of the view if resolved, or the original path otherwise
 */
function resolvePath(path) {
    var normalizedPath = path;
    if (normalizedPath.endsWith(".njk")) {
        normalizedPath = normalizedPath.substring(0, normalizedPath.length - 4);
    }
    var app = this.ctx.req.app.get("app");
    var resolved = app.templateMap[path];
    if (resolved) {
        return resolved;
    }
    return path;
}
/**
 * The App class contains the initialization code for a Lynx application.
 */
var App = /** @class */ (function () {
    function App(config, modules) {
        this._config = config;
        if (modules) {
            for (var _i = 0, modules_1 = modules; _i < modules_1.length; _i++) {
                var module_1 = modules_1[_i];
                module_1.mount(this._config);
            }
        }
        config.db.entities.unshift(__dirname + "/entities/*.entity.js");
        config.middlewaresFolders.unshift(__dirname + "/middlewares");
        config.viewFolders.unshift(__dirname + "/views");
        if (!config.disabledDb) {
            typeorm_1.createConnection(config.db)
                .then(function (_) {
                // here you can start to work with your entities
                console.log("Connection to the db established!");
                setup_1.setup(config.db.entities).catch(function (error) {
                    console.log(error);
                    process.exit(1);
                });
            })
                .catch(function (error) {
                console.error(error);
                process.exit(1);
            });
        }
        else {
            console.log("The DB service is disabled");
        }
        this.express = express();
        this.express.set("app", this);
        this.express.use("/api/*", cors());
        this.express.use(bodyParser.json());
        this.express.use(bodyParser.urlencoded({ extended: true }));
        var app_session_options = {
            secret: config.sessionSecret,
            resave: false,
            saveUninitialized: true,
            cookie: {}
        };
        if (isProduction()) {
            this.express.set("trust proxy", 1);
            app_session_options.cookie.secure = true;
        }
        if (config.sessionStore) {
            app_session_options.store = config.sessionStore;
        }
        var app_session = session(app_session_options);
        this.express.use(app_session);
        this._upload = multer({ dest: config.uploadPath });
        for (var _a = 0, _b = config.publicFolders; _a < _b.length; _a++) {
            var folder = _b[_a];
            this.express.use(express.static(folder));
        }
        this.generateTemplateMap(config.viewFolders);
        this._nunjucksEnvironment = nunjucks.configure(config.viewFolders, {
            autoescape: true,
            watch: true,
            express: this.express
        });
        this._nunjucksEnvironment.addFilter("tr", translate);
        this._nunjucksEnvironment.addFilter("json", JSON.stringify);
        this._nunjucksEnvironment.addFilter("format", format);
        this.loadTranslations(config.translationFolders);
        this._nunjucksEnvironment.addGlobal("route", route);
        this._nunjucksEnvironment.addGlobal("old", old);
        this._nunjucksEnvironment.addGlobal("resolvePath", resolvePath);
        for (var _c = 0, _d = config.middlewaresFolders; _c < _d.length; _c++) {
            var path = _d[_c];
            this.loadMiddlewares(path);
        }
        for (var _e = 0, _f = config.controllersFolders; _e < _f.length; _e++) {
            var path = _f[_e];
            this.loadControllers(path);
        }
        if (!config.disabledDb) {
            var schema = graphqlGenerator.generateSchema(config.db.entities);
            // The GraphQL endpoint
            this.express.use("/graphql", bodyParser.json(), apollo_server_express_1.graphqlExpress({ schema: schema }));
            // GraphiQL, a visual editor for queries
            this.express.use("/graphiql", apollo_server_express_1.graphiqlExpress({ endpointURL: "/graphql" }));
        }
    }
    Object.defineProperty(App.prototype, "config", {
        get: function () {
            return this._config;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(App.prototype, "templateMap", {
        get: function () {
            return this._templateMap;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(App.prototype, "nunjucksEnvironment", {
        get: function () {
            return this._nunjucksEnvironment;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(App.prototype, "upload", {
        get: function () {
            return this._upload;
        },
        enumerable: true,
        configurable: true
    });
    App.prototype.recursiveGenerateTemplateMap = function (path, currentPath) {
        var files = fs.readdirSync(path);
        for (var index in files) {
            var currentFilePath = path + "/" + files[index];
            if (fs.lstatSync(currentFilePath).isDirectory()) {
                this.recursiveGenerateTemplateMap(currentFilePath, currentPath + files[index] + "/");
                continue;
            }
            var name = files[index].replace(".njk", "");
            this._templateMap[currentPath + name] = currentFilePath;
        }
    };
    App.prototype.generateTemplateMap = function (paths) {
        this._templateMap = {};
        for (var _i = 0, paths_1 = paths; _i < paths_1.length; _i++) {
            var path = paths_1[_i];
            this.recursiveGenerateTemplateMap(path, "/");
        }
    };
    App.prototype.loadTranslations = function (paths) {
        for (var _i = 0, paths_2 = paths; _i < paths_2.length; _i++) {
            var path = paths_2[_i];
            var files = fs.readdirSync(path);
            for (var index in files) {
                var nameWithExtension = files[index];
                if (!nameWithExtension.endsWith("json"))
                    continue;
                var name = nameWithExtension.substring(0, nameWithExtension.indexOf("."));
                var tmp = JSON.parse(fs.readFileSync(path + "/" + nameWithExtension, "utf8"));
                if (!translations[name]) {
                    translations[name] = {};
                }
                for (var key in tmp) {
                    translations[name][key] = tmp[key];
                }
            }
        }
    };
    App.prototype.loadMiddlewares = function (path) {
        var middlewares = fs.readdirSync(path);
        for (var index in middlewares) {
            var currentFilePath = path + "/" + middlewares[index];
            if (fs.lstatSync(currentFilePath).isDirectory()) {
                this.loadMiddlewares(currentFilePath);
                continue;
            }
            if (middlewares[index].endsWith("ts"))
                continue;
            var midd = require(currentFilePath);
            if (!midd.default) {
                throw new Error("Plese define the middleware as the export default class in file " +
                    currentFilePath +
                    ".");
            }
            expressGenerator.useMiddleware(this, midd.default);
        }
    };
    App.prototype.loadControllers = function (path) {
        var files = fs.readdirSync(path);
        for (var index in files) {
            var currentFilePath = path + "/" + files[index];
            if (fs.lstatSync(currentFilePath).isDirectory()) {
                this.loadControllers(currentFilePath);
                continue;
            }
            if (files[index].endsWith("ts"))
                continue;
            var ctrl = require(currentFilePath);
            if (!ctrl.default) {
                throw new Error("Plese define the controller as the export default class in file " +
                    currentFilePath +
                    ".");
            }
            expressGenerator.useController(this, ctrl.default, routes);
        }
    };
    App.prototype.startServer = function (port) {
        this.express.listen(port, function (err) {
            if (err) {
                return console.log(err);
            }
            return console.log("server is listening on " + port);
        });
    };
    App.prototype.route = function (name, parameters) {
        return route(name, parameters);
    };
    App.prototype.translate = function (str, req) {
        try {
            var lang = this._config.defaultLanguage;
            var langs = req.acceptsLanguages();
            if (langs && langs.length > 0) {
                lang = langs[0];
            }
            return performTranslation(str, translations[lang]);
        }
        catch (e) {
            console.log(e);
        }
        return str;
    };
    App.prototype.generateTokenForUser = function (user) {
        return jsonwebtoken_1.sign({ id: user.id }, this._config.tokenSecret, {
            expiresIn: "1y"
        });
    };
    return App;
}());
exports.default = App;
