"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var http_verb_1 = require("./http-verb");
var routes = [];
/**
 * Decorator to set the base route of a controller.
 * @param path the route path for the controller.
 */
function Route(path) {
    return function (target) {
        target.controllerPath = path;
        target.routes = routes;
        routes = [];
    };
}
exports.Route = Route;
function setTarget(_, type, path, method) {
    routes.push({
        type: type,
        path: path,
        method: method,
        body: undefined,
        isAPI: false,
        isMultipartForm: false,
        verifiers: []
    });
}
function setBody(_, name, schema) {
    if (!routes)
        return;
    routes[routes.length - 1].body = {
        name: name,
        schema: schema
    };
}
function setApi(_) {
    if (!routes)
        return;
    routes[routes.length - 1].isAPI = true;
}
function setMultipartForm(_) {
    if (!routes)
        return;
    routes[routes.length - 1].isMultipartForm = true;
}
function setName(name) {
    if (!routes)
        return;
    routes[routes.length - 1].name = name;
}
function setVerify(_, func) {
    if (!routes)
        return;
    routes[routes.length - 1].verifiers.push(func);
}
/**
 * Set the decorated method to a GET endpoint with the specified path.
 * @param path the endpoint path
 */
function GET(path) {
    return function (target, method, _) {
        setTarget(target, http_verb_1.HttpVerb.GET, path, method);
    };
}
exports.GET = GET;
/**
 * Set the decorated method to a POST endpoint with the specified path.
 * @param path the endpoint path
 */
function POST(path) {
    return function (target, method, _) {
        setTarget(target, http_verb_1.HttpVerb.POST, path, method);
    };
}
exports.POST = POST;
/**
 * Set the decorated method to a PUT endpoint with the specified path.
 * @param path the endpoint path
 */
function PUT(path) {
    return function (target, method, _) {
        setTarget(target, http_verb_1.HttpVerb.PUT, path, method);
    };
}
exports.PUT = PUT;
/**
 * Set the decorated method to a DELETE endpoint with the specified path.
 * @param path the endpoint path
 */
function DELETE(path) {
    return function (target, method, _) {
        setTarget(target, http_verb_1.HttpVerb.DELETE, path, method);
    };
}
exports.DELETE = DELETE;
/**
 * Set the decorated method to a PATH endpoint with the specified path.
 * @param path the endpoint path
 */
function PATCH(path) {
    return function (target, method, _) {
        setTarget(target, http_verb_1.HttpVerb.PATCH, path, method);
    };
}
exports.PATCH = PATCH;
/**
 * Add to the decorated method a body to be injected. The body will be validated
 * using the specified schema.
 * @param name the name of the argument in the method to map the body
 * @param schema a JOI schema to validate the body object
 */
function Body(name, schema) {
    return function (target, _, __) {
        setBody(target, name, schema);
    };
}
exports.Body = Body;
/**
 * Set the decorated method to an API endpoints.
 * In this way, the returned value of the method will be encapsulated in a
 * standard API envelope and serialized to JSON using the Lynx serialization system.
 */
function API() {
    return function (target, _, __) {
        setApi(target);
    };
}
exports.API = API;
/**
 * Set the decorated method to accept MultipartForm requested.
 */
function MultipartForm() {
    return function (target, _, __) {
        setMultipartForm(target);
    };
}
exports.MultipartForm = MultipartForm;
/**
 * Add to the decorated method a route name, in order to easely generate redirect
 * or, more general, the execution of the `route` method.
 * @param name the name of the route
 */
function Name(name) {
    return function (_, __, ___) {
        setName(name);
    };
}
exports.Name = Name;
/**
 * Add to the decorated method a verification function that will be executed
 * BEFORE the route. The function must NOT be an async function, and it shell
 * return a boolean value. If true is returned, the method is then executed.
 * This method is fundamental to implement authorization to a single endpoint.
 * @param func the verification function to be executed. It must NOT be an async function, and return a boolean value.
 */
function Verify(func) {
    return function (target, _, __) {
        setVerify(target, func);
    };
}
exports.Verify = Verify;
/**
 * Add to the decorated class a path to be executed as middleware.
 * @param path the endpoint path
 */
function Middleware(path) {
    return function (target) {
        target.middlewarePath = path;
    };
}
exports.Middleware = Middleware;
